#!/usr/bin/perl

use strict;
use IO::File;
use IO::Socket qw(:DEFAULT :crlf);

use constant DSDIR        => '/opt/ds';
use constant DSVAR        => '/var/ds';
use constant SERVER_PORT  => 8511;
use constant DS_CONFIG    => DSDIR.'/etc/dserv.conf';
use constant DSLOGD_LOG   => DSVAR.'/log/dslogd.log';
use constant MAX_LOG_SIZE => 1 << 20; # 1 Mb
use constant BUFFER_SIZE  => 1 << 20; # 1 Mb

$/ = CRLF;

$SIG{PIPE} = 'IGNORE';
$SIG{__WARN__} = \&log_warn;
$SIG{__DIE__}  = \&log_die;

my $fh;
sub chkLog {
    if(!defined($fh) or not -e DSLOGD_LOG) {
	$fh = IO::File->new(DSLOGD_LOG, O_WRONLY|O_APPEND|O_CREAT, 0644);
	$fh->autoflush(1) if defined $fh;
    }
    return undef if not defined $fh or not -e DSLOGD_LOG;
    my $logFile = DSLOGD_LOG;
    my $logSize = -s DSLOGD_LOG;
    my ($record_len) = @_;
    if($logSize + length($record_len) + 2 > MAX_LOG_SIZE) {
	close($fh);
	`mv -f $logFile $logFile.bak`;
	`rm -f $logFile`;
	$fh = IO::File->new(DSLOGD_LOG, O_WRONLY|O_APPEND|O_CREAT, 0644);
	$fh->autoflush(1) if defined $fh;
    }
    return $fh;
}

sub log_info  { print "[INFO] @_\n"; print $fh localtime()." [INFO] @_\n" if defined chkLog(@_); }
sub log_warn  { print "[WARN] @_\n"; print $fh localtime()." [WARN] @_\n" if defined chkLog(@_); }
sub log_die   { print $fh localtime()." [FATAL] @_\n" if defined chkLog(@_); die @_; }


# Parse DS_CONFIG
my ($cfg, $usr, $md5) = DS_CONFIG;
if(`cat $cfg` =~ /<USER\s+ID=\"(.+)\"\s+MD5=\"([\w\d]+)\"/i) {
    $usr = $1;
    $md5 = $2;
}


my $server_port = shift || SERVER_PORT;
my $sock = IO::Socket::INET->new(LocalPort => $server_port,
                                 Type      => SOCK_STREAM,
                                 Reuse     => 1,
                                 Listen    => SOMAXCONN )
    or die "Can't create listening socket on port $server_port : $@\n";

my $client;
log_info "Server is started. Waiting for connections...";
while ($client = $sock->accept()) {
    my $peer = gethostbyaddr($client->peeraddr, AF_INET) || $client->peerhost;
    my $port = $client->peerport;
    log_info "Connection from [$peer:$port]";

    # receiving request from client
    my ($req, $type, $uni, $user, $pass_md5) = ('', '', '', '', '');
    while (<$client>) {
	$req .= $_;
	$type = $1 if /^GET (LOG|STAT)\s*/i;
	$uni  = $1 if /^UNI: (.+)\s*/i;
	$user = $1 if /^USRNAME: (.+)\s*/i;
	$pass_md5 = $1 if /^MD5: (.+)\s*/i;
	if(/^\s+/) {
	    log_info "CLIENT REQUEST:\n".$req;
	    # sending response to client
	    if($user != $usr || $pass_md5 != $md5) {
		# send response with error status
		my $response = "DSLOGD$/STATUS: ERROR$/MESSAGE: Invalid USRNAME or PASSWORD: $user:$pass_md5$/$/";
		print $client $response;
		warn "SERVER RESPONSE:\n".$response."Closing client connection\n\n";
		close($client);
		last;
	    } elsif ($type eq 'STAT') {
		my $content = '';
		my $tmp = "\n\n\n";
		my $gdb_commands = '/tmp/gdb_commands';
		my @arr = (
		    'df -h',
		    'uname -a',
		    'cat /etc/redhat-release',
		    'rpm -qa|grep ds-vmx',
		    'cat '.DSDIR.'/etc/dserv.conf',
		    'cat /etc/X11/xorg.conf',
		    'ifconfig -a',
		    'top -n 1',
		    'ps aux',
		    "ps -eo pid,pmem,pcpu,rss,etime,cmd|grep \"\\.\\./bin/DS\"|grep -v grep",
		    "ps -eLf|grep \"\\.\\./bin/DS\"",
		    "ps -eLf|grep \"\\.\\./bin/DS\"|wc -l",
		    'netstat',
		    'netstat|grep 8510',
		    'netstat|grep 8510|wc -l',
		    'netstat|grep acpid|wc -l',
		    'xvinfo -display :0',
		    'xdpyinfo -display :0',
		    "ps -eo pid,cmd|grep \"\\.\\./bin/DS\"|grep -v grep|cut -c 1-6|xargs pmap -x",
		    "ps -eo pid,cmd|grep \"\\.\\./bin/DS\"|grep -v grep|cut -c 1-6|xargs gdb -x $gdb_commands -p");

		my ($count, $thr_num) = (0, `ps -eLf|grep \"\\.\\./bin/DS\"|wc -l`);
		`echo "info threads\n" > $gdb_commands`;
		while(++$count <= $thr_num) {
		    `echo "thread $count\nbt\n" >> $gdb_commands`;
		}
		`echo "q\n" >> $gdb_commands`;

		foreach my $cmd (@arr) {
		    $content .= "# $cmd\n".`$cmd`.$tmp;
		}
		`rm -f $gdb_commands`;
		
		my $size = length($content);
		print 'Content size to send: '.$size."\n";
		my $response = "DSLOGD$/STATUS: OK$/CONTENT-LENGTH: $size$/FILE: DS_stat.log$/$/$content";
		print $client $response;
		log_info "SERVER RESPONSE:\n".$response;
	    } elsif ($type eq 'LOG') {
		my $filename = '/tmp/DS_logs.tgz';
		my $tmp = DSVAR.'/log';
		`cd $tmp && tar -czf $filename *`;
		if(-e $filename) {
		    my $size = -s $filename;
		    my $response = "DSLOGD$/STATUS: OK$/CONTENT-LENGTH: $size$/FILE: DS_logs.tgz$/$/";
		    log_info "SERVER RESPONSE:\n".$response."..........\n\n";
		    print $client $response;
		    
		    my ($Size, $chunk_size, $bytes_read, $bytes_written) = (0, 0, 0, 0);
		    my $buffer = "";
		    open(F, "<$filename") or die "Cannot open file for reading: $!";
		    while($Size < $size) {
			$chunk_size = $size - $Size < BUFFER_SIZE ? $size - $Size : BUFFER_SIZE;
			$bytes_read = read(F, $buffer, $chunk_size);
			die "Error occured while reading data from file $filename: $!" unless defined $bytes_read;
			die "End of file occured ($filename)" unless $bytes_read;
			$bytes_written = syswrite($client, $buffer);
			warn("Error occured while transferring data from file $filename: $!"),last unless defined $bytes_written;
			warn("Zero bytes transfered from file $filename"),last unless $bytes_written;
			warn("Bytes written not equal to bytes read from file $filename"),last if $bytes_written != $bytes_read;
			$Size += $bytes_written;
		    }
		    close F;
		    `rm -f $filename`;
		}
		else {
		    my $response = "DSLOGD$/STATUS: ERROR$/MESSAGE: Failed to create archive$/";
		    print $client $response;
		    warn "SERVER RESPONSE:\n".$response."\n";
		}
	    } else {
		my $response = "DSLOGD$/STATUS: ERROR$/MESSAGE: Command $type not supported$/$/";
		print $client $response;
		warn "SERVER RESPONSE:\n".$response;
	    }

	    ($req, $type, $uni, $user, $pass_md5) = ('', '', '', '', '');
	    close($client);
	}
    }
    log_info "Waiting for connections...";
}

close($sock);
warn "Server has finished the work! THIS MESSAGE MUST NEVER BE PRINTED!";
